/*
 * Copyright (C) 2015 Jolla Ltd.
 * Contact: Jonni Rainisto <jonni.rainisto@jollamobile.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef DROPBOXIMAGEDOWNLOADER_P_H
#define DROPBOXIMAGEDOWNLOADER_P_H

#include <QtCore/QObject>
#include <QtCore/QThread>
#include <QtCore/QMutex>
#include <QtCore/QWaitCondition>
#include <QtCore/QSet>

#include "abstractimagedownloader_p.h"
#include "dropboximagedownloader.h"
#include "dropboximagesdatabase.h"

class DropboxImageCacheModel;
class DropboxImageDownloaderPrivate: public AbstractImageDownloaderPrivate
{
public:
    explicit DropboxImageDownloaderPrivate(DropboxImageDownloader *q);
    virtual ~DropboxImageDownloaderPrivate();

    DropboxImagesDatabase database;
    QSet<DropboxImageCacheModel*> m_connectedModels;

private:
    Q_DECLARE_PUBLIC(DropboxImageDownloader)
};

#endif // DROPBOXIMAGEDOWNLOADER_P_H
